"""A modification workflow for a site."""

from orchestrator.forms import FormPage
from orchestrator.targets import Target
from orchestrator.types import FormGenerator, State, SubscriptionLifecycle, UUIDstr
from orchestrator.workflow import StepList, begin, done, step, workflow
from orchestrator.workflows.steps import (
    resync,
    set_status,
    store_process_subscription,
    unsync,
)
from orchestrator.workflows.utils import wrap_modify_initial_input_form
from pydantic import ConfigDict, field_validator
from pydantic_core.core_schema import ValidationInfo
from pydantic_forms.validators import ReadOnlyField

from gso.products.product_blocks.site import LatitudeCoordinate, LongitudeCoordinate, SiteTier
from gso.products.product_types.site import Site
from gso.utils.helpers import validate_ipv4_or_ipv6, validate_site_fields_is_unique


def initial_input_form_generator(subscription_id: UUIDstr) -> FormGenerator:
    """Gather input from the operator on what to change about the selected site subscription."""
    subscription = Site.from_subscription(subscription_id)

    class ModifySiteForm(FormPage):
        model_config = ConfigDict(title="Modify Site")

        site_name: ReadOnlyField(subscription.site.site_name, default_type=str)  # type: ignore[valid-type]
        site_city: str = subscription.site.site_city
        site_country: ReadOnlyField(subscription.site.site_country, default_type=str)  # type: ignore[valid-type]
        site_country_code: ReadOnlyField(subscription.site.site_country_code, default_type=str)  # type: ignore[valid-type]
        site_latitude: LatitudeCoordinate = subscription.site.site_latitude
        site_longitude: LongitudeCoordinate = subscription.site.site_longitude
        site_bgp_community_id: int = subscription.site.site_bgp_community_id
        site_internal_id: int = subscription.site.site_internal_id
        site_tier: ReadOnlyField(subscription.site.site_tier, default_type=SiteTier)  # type: ignore[valid-type]
        site_ts_address: str | None = subscription.site.site_ts_address

        @field_validator("site_ts_address", "site_internal_id", "site_bgp_community_id")
        def field_must_be_unique(cls, value: str | int, info: ValidationInfo) -> str | int:
            if not info.field_name:
                msg = "Field name must be provided."
                raise ValueError(msg)

            if value and value == getattr(subscription.site, info.field_name):
                return value

            validate_site_fields_is_unique(info.field_name, value)

            return value

        @field_validator("site_ts_address")
        def validate_ts_address(cls, site_ts_address: str) -> str:
            if site_ts_address and site_ts_address != subscription.site.site_ts_address:
                validate_ipv4_or_ipv6(site_ts_address)

            return site_ts_address

    user_input = yield ModifySiteForm

    return user_input.dict()


@step("Modify subscription")
def modify_site_subscription(
    subscription: Site,
    site_city: str,
    site_latitude: LatitudeCoordinate,
    site_longitude: LongitudeCoordinate,
    site_bgp_community_id: int,
    site_internal_id: int,
    site_ts_address: str,
) -> State:
    """Update the subscription model in the service database."""
    subscription.site.site_city = site_city
    subscription.site.site_latitude = site_latitude
    subscription.site.site_longitude = site_longitude
    subscription.site.site_bgp_community_id = site_bgp_community_id
    subscription.site.site_internal_id = site_internal_id
    subscription.site.site_ts_address = site_ts_address

    subscription.description = f"Site in {site_city}, {subscription.site.site_country}"

    return {"subscription": subscription}


@workflow(
    "Modify Site",
    initial_input_form=wrap_modify_initial_input_form(initial_input_form_generator),
    target=Target.MODIFY,
)
def modify_site() -> StepList:
    """Modify a site subscription.

    * Update the subscription model in the service database
    """
    return (
        begin
        >> store_process_subscription(Target.MODIFY)
        >> unsync
        >> modify_site_subscription
        >> set_status(SubscriptionLifecycle.ACTIVE)
        >> resync
        >> done
    )
